/******************************************************************************
 * Icinga 2                                                                   *
 * Copyright (C) 2012 Icinga Development Team (http://www.icinga.org/)        *
 *                                                                            *
 * This program is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU General Public License                *
 * as published by the Free Software Foundation; either version 2             *
 * of the License, or (at your option) any later version.                     *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program; if not, write to the Free Software Foundation     *
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.             *
 ******************************************************************************/

#ifndef HOST_H
#define HOST_H

namespace icinga
{

class Service;

/**
 * The state of a host.
 *
 * @ingroup icinga
 */
enum HostState
{
	HostUp = 0,
	HostDown = 1,
	HostUnreachable = 2
};

/**
 * The state type of a host or service.
 *
 * @ingroup icinga
 */
enum StateType
{
	StateTypeSoft = 0,
	StateTypeHard = 1
};

/**
 * An Icinga host.
 *
 * @ingroup icinga
 */
class I2_ICINGA_API Host : public DynamicObject
{
public:
	typedef shared_ptr<Host> Ptr;
	typedef weak_ptr<Host> WeakPtr;

	Host(const Dictionary::Ptr& properties);
	~Host(void);

	static Host::Ptr GetByName(const String& name);

	String GetDisplayName(void) const;
	Dictionary::Ptr GetGroups(void) const;

	Dictionary::Ptr GetMacros(void) const;
	Dictionary::Ptr GetHostDependencies(void) const;
	Dictionary::Ptr GetServiceDependencies(void) const;
	String GetHostCheck(void) const;

	Dictionary::Ptr CalculateDynamicMacros(void) const;

	shared_ptr<Service> GetHostCheckService(void) const;
	set<Host::Ptr> GetParentHosts(void) const;
	set<shared_ptr<Service> > GetParentServices(void) const;

	bool IsReachable() const;

	shared_ptr<Service> GetServiceByShortName(const Value& name) const;

	set<shared_ptr<Service> > GetServices(void) const;
	static void InvalidateServicesCache(void);

	static void ValidateServiceDictionary(const ScriptTask::Ptr& task,
	    const std::vector<icinga::Value>& arguments);

	HostState GetState(void) const;
	StateType GetStateType(void) const;

	HostState GetLastState(void) const;
	StateType GetLastStateType(void) const;

	static String HostStateToString(HostState state);

protected:
	virtual void OnRegistrationCompleted(void);
	virtual void OnAttributeChanged(const String& name);

private:
	Attribute<String> m_DisplayName;
	Attribute<Dictionary::Ptr> m_HostGroups;
	Attribute<Dictionary::Ptr> m_Macros;
	Attribute<Dictionary::Ptr> m_HostDependencies;
	Attribute<Dictionary::Ptr> m_ServiceDependencies;
	Attribute<String> m_HostCheck;
	Dictionary::Ptr m_SlaveServices;

	static boost::mutex m_ServiceMutex;
	static map<String, map<String, weak_ptr<Service> > > m_ServicesCache;
	static bool m_ServicesCacheNeedsUpdate;
	static Timer::Ptr m_ServicesCacheTimer;

	void UpdateSlaveServices(void);

	static void RefreshServicesCache(void);
};

}

#endif /* HOST_H */
